`default_nettype none
module sigmadelta(x, norot, SevenofNine, sdclock, sdclock5, rnd1, rnd2, rot, y, iclip, scale, en);
// Vereenvoudigde versie die alleen PBM8 doet, met of zonder willekeurige rotatie, met 7 of 9 toegestane kwantisatorniveaus.
// Willekeurige rotatie aan en 9 niveaus is het beste voor een RTZ-DAC, een NRZ-DAC werkt mogelijk beter met de willekeurige
// rotatie uit en 7 kwantisatieniveaus, want dan heb je altijd precies 1 stijgende en 1 dalende flank per groep van acht bits.
// Simplified version that only does PWM8, with or without random rotation, with 7 or 9 allowed quantization levels.
// Random rotation on and 9 quantization levels is best for an RTZ DAC, random rotation off and 7 quantization levels may
// work better with an NRZ DAC, because you then always get precisely one rising and one falling edge per group of eight bits.

// norot = 1 -> willekeurige rotatie uit, random rotation off
// SevenofNine = 1 -> Annika Hansen-modus, maar zeven van de negen kwantisatieniveaus toegestaan, only seven out of the nine quantization levels allowed

// scale:
// 0: ingang 39/64 volle schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 0: 39/64 of full-scale input drives the sigma-delta to 75 % ones / 25 % ones
// 1: ingang 39/64 halve schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 1: 39/64 of half-scale input drives the sigma-delta to 75 % ones / 25 % ones
// 2: ingang 39/64 kwart schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 2: 39/64 of quarter-scale input drives the sigma-delta to 75 % ones / 25 % ones
// 3: ingang 39/64 achtste schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 3: 39/64 of eighth-scale input drives the sigma-delta to 75 % ones / 25 % ones
// Die factor 39/64 heeft te maken met de terugkoppelcoefficient, waarvoor in het 
// interpolatiefilter gecorrigeerd wordt om een vermenigvuldiger uit te sparen.
// The factor 39/64 is related to the feedback coefficient, which is corrected
// for in the interpolation filter to save a multiplier.

// en: enable, indien 0 wordt de interne activiteit beperkt en worden sommige toestandsvariabelen op 0 gehouden
// en: enable, if 0 the internal activity is reduced and some state variables are forced to 0

parameter fractbits = 29;
// aantal bits achter de (fictieve) komma na afronding
// number of bits after the (fictituous) binary point after truncation

wire signed [13:0] ceff[7:0]; // coefficienten afgerond op 13 bits achter de komma, coefficients rounded to 13 bits after the binary point
// Het 14de bit is altijd 1 vanwege de manier waarop het tegengestelde berekend wordt wanneer de sigma-deltauitgang 0 is. 
// The 14th bit is always 1 due to the way the opposite is calculated when the sigma-delta output is 0.
assign ceff[0] = 14'sd19; // eigenlijk/actually 19.5/8192 = 39/16384
assign ceff[1] = 14'sd216; // eigenlijk/actually 216.5/8192 = 433/16384
assign ceff[2] = 14'sd1036; // eigenlijk/actually 1036.5/8192 = 2073/16384
assign ceff[3] = 14'sd3047; // eigenlijk/actually 3047.5/8192 = 6095/16384
assign ceff[4] = 14'sd5142; // eigenlijk/actually 5142.5/8192 = 10285/16384
assign ceff[5] = 14'sd0;
assign ceff[6] = 14'sd0;
assign ceff[7] = 14'sd0;

parameter wordlengthx = fractbits + 14 - 8;

input wire signed [wordlengthx-1:0] x; // x heeft fractbits + 14 cijfers achter de komma, x has fractbits + 14 digits after the binary point
//
// De bits voor en de eerste acht achter de komma zijn niet van belang vanwege de schaling. Met scale = 0 is +/- 39/32768 al voldoende om de 
// modulator uit te sturen van 1/4 enen tot 3/4 enen, met scale = 1, 2 of 3 wordt de ingang nog gevoeliger.
//
// The bits before and the first eight bits after the binary point are irrelevant because of the scaling. With scale = 0, +/- 39/32768 suffices
// to drive the modulator from 1/4 to 3/4 ones, with scale = 1, 2 or 3 the input gets more sensitive.

input wire norot;
input wire SevenofNine;
input wire sdclock, sdclock5; // sdclock5: synchroon met sdclock, vijf keer de frequentie, synchronous with sdclock, but five times as high frequency

input wire [fractbits+13:0] rnd1; // expres zonder teken, deliberately unsigned
input wire signed [fractbits+13:0] rnd2;
// rnd1 en rnd2 worden gebruikt voor het ditheren, zowel om de woordlengtes eindig te houden als voor de (hoofd-) kwantisator.
// rnd1 and rnd2 are used for dithering, both to keep the wordlengths finite and for the (main) quantizer

input wire [2:0] rot;
input wire [1:0] scale;
output wire y;
output reg iclip;
input wire en;
reg [3:0] q;
reg signed [wordlengthx+2:0] xsh; // Als je x drie plaatsen naar links schuift, moet het nog in xsh passen. When you shift x three places to the left, it still needs to fit in xsh. 
reg signed [fractbits+14+4:0] shiftreg1 = {(fractbits+14+5) {1'b0}};
reg signed [fractbits+14+4:0] newval;
reg signed [fractbits+14+4:0] newval1;
reg signed [fractbits+4:0] shiftreg2 = {(fractbits+5) {1'b0}};
reg signed [fractbits+3:0] shiftreg2lim;
reg signed [fractbits+3:0] shiftreg3 = {(fractbits+4) {1'b0}};
reg signed [fractbits+3:0] shiftreg4 = {(fractbits+4) {1'b0}};
reg signed [fractbits+3:0] shiftreg5 = {(fractbits+4) {1'b0}};
reg signed [fractbits+3:0] shiftreg6 = {(fractbits+4) {1'b0}};

reg signed [fractbits+14+4:0] partdith;
reg signed [fractbits+4:0] roundedfine;
reg signed [fractbits+4:0] quantintermediate;

reg ol8p, ol4p, ol1p, olquarterp, ol8n, ol4n, ol1n, olquartern, olp, oln;
reg signed [fractbits+3:0] ollp;
reg signed [fractbits+3:0] olln;

reg [2:0] shiftcount = 3'b000;

reg [2:0] rotifen;

reg Piet = 1'b0;
reg Pietd = 1'b0;

always @ (*)
begin
	if (norot == 1'b0)
		rotifen = rot; // rot if random rotation is enabled, else zero
	else
		rotifen = 3'b000;
end

// Het volgende is een trucje om transities van sdclock te kunnen detecteren zonder dat sdclock via de normale bedrading
// hoeft te gaan (sdclock gaat gewoon via de boomstructuren voor de klokken).
// This is a trick to detect transitions of sdclock without the synthesizer having to route it via ordinary wires, rather
// than the tree structures meant for clocks.
always @ (posedge sdclock)
begin
		Piet <= !Piet;
end

always @ (posedge sdclock5)
begin
	if (en)
	begin
		shiftreg6 <= shiftreg5;
		shiftreg5 <= shiftreg4;
		shiftreg4 <= shiftreg3;
		shiftreg3 <= shiftreg2lim; // toestandsvariabelebegrenzing, state variable limiting
	end
	else
	begin
		shiftreg6 <= {(fractbits+4) {1'b0}};
		shiftreg5 <= {(fractbits+4) {1'b0}};
		shiftreg4 <= {(fractbits+4) {1'b0}}; // toestandsvariabelen op 0 gezet, indien mogelijk zonder timingproblemen
		shiftreg3 <= {(fractbits+4) {1'b0}}; // state variables set to 0, where possible without timing issues
	end
	shiftreg2 <= roundedfine; // geditherde afronding, dithered truncation
	shiftreg1 <= newval;
// newval en shiftreg1 hebben (fractbits + 14) bits achter de komma, shiftreg2...shiftreg6 fractbits bits achter de komma
// newval and shiftreg1 have (fractbits + 14) bits after the binary point, shiftreg2...shiftreg6 fractbits bits after the binary point
// newval, shiftreg1 en shiftreg2 hebben 5 bits voor de komma (inclusief tekenbit), shiftreg3...shiftreg6 4 bits voor de komma (inclusief tekenbit)
// newval, shiftreg1 and shiftreg2 have 5 bits before the binary point (sign bit included), shiftreg3...shiftreg6 4 bits before the binary point (sign bit included)
	Pietd <= Piet;
	if (Piet ^^ Pietd)
		shiftcount <= 3'b010;
	else
	begin
		if (shiftcount[2] == 1)
			shiftcount <= 3'b000;
		else
			shiftcount <= shiftcount + 3'b001;
	end
end
always @ (*)
begin
	case (scale)
		2'b00: xsh = x; // x moet het juiste aantal cijfers achter de komma hebben! x must have the right number of bits after the binary point!
		2'b01: xsh = (x <<< 1);
		2'b10: xsh = (x <<< 2);
		2'b11: xsh = (x <<< 3);
	endcase
	case (shiftcount)
		3'b000: newval1 = (shiftreg5 <<< 14) + xsh - shiftreg4;
		3'b010: newval1 = (shiftreg5 <<< 14) + (shiftreg6 <<< 11) - (shiftreg4 <<< 4);
		default: newval1 = (shiftreg5 <<< 14) + (shiftreg6 <<< 11);
	endcase

// Tegenkoppeling: ceff aftrekken als het uitgangsbit 1 is, anders optellen. Tegengestelde van ceff benaderd als inverse. Offset gecorrigeerd.
// Negative feedback: subtract ceff when the output bit is 1, otherwise add it. Opposite approximated with the inverse. Resulting offset corrected.
	if (y == 1'b1)
		newval = newval1 + $signed({(~ceff[shiftcount]), 1'b1, {(fractbits-14+14) {1'b0}}});
	else
		newval = newval1 + $signed({ceff[shiftcount], 1'b1, {(fractbits-14+14) {1'b0}}});
// De ceff's hebben 13 bits achter de komma, newval heeft er fractbits + 14. Aanvullen met een 1 (offsetcorrectie) en fractbits nullen dus.
// The ceff's have 13 bits after the binary point, newval has fractbits + 14. Hence, extend with a one (offset correction) and fractbits zeros.

	partdith = $unsigned(shiftreg1) + $unsigned(rnd1[13:0]) + $unsigned(rnd2[13:0]); // partdith heeft desalniettemin een teken, partdith is nonetheless signed
	roundedfine = (partdith >>> 14);
//	quantintermediate = roundedfine + $signed({1'b0, rnd1[fractbits+13:14]})+ $signed(rnd2[fractbits+13:14]);
	quantintermediate = (shiftreg1 + $signed({1'b0, rnd1[fractbits+13:0]})+ $signed(rnd2[fractbits+13:0])) >>> 14;
	casex ({quantintermediate[fractbits+4:fractbits], SevenofNine})
		6'b10xxx0: q = 4'b0000;
		6'b10xxx1: q = 4'b0001;
		6'b110xx0: q = 4'b0000;
		6'b110xx1: q = 4'b0001;
		6'b111000: q = 4'b0000;
		6'b111001: q = 4'b0001;
		6'b11101x: q = 4'b0001;
		6'b11110x: q = 4'b0010;
		6'b11111x: q = 4'b0011;
		6'b00000x: q = 4'b0100;
		6'b00001x: q = 4'b0101;
		6'b00010x: q = 4'b0110;
		6'b00011x: q = 4'b0111;
		6'b001xx0: q = 4'b1000;
		6'b001xx1: q = 4'b0111;
		6'b01xxx0: q = 4'b1000;
		6'b01xxx1: q = 4'b0111;
	endcase

// Toestandsvariabelebegrenzing, we kunnen wegkomen met machten van twee als grenzen
// State variable limiting, we can get away with using powers of two as limits
	ol8p = (!shiftreg2[fractbits+4]) && shiftreg2[fractbits+3]; // shiftreg2 >= 8
	ol4p = (!shiftreg2[fractbits+4]) && (|shiftreg2[fractbits+3:fractbits+2]); // shiftreg2 >= 4
	ol1p = (!shiftreg2[fractbits+4]) && (|shiftreg2[fractbits+3:fractbits]); // shiftreg2 >= 1
	olquarterp = (!shiftreg2[fractbits+4]) && (|shiftreg2[fractbits+3:fractbits-2]); // shiftreg2 >= 1/4
	ol8n = (shiftreg2[fractbits+4]) && (!shiftreg2[fractbits+3]); // shiftreg2 < -8
	ol4n = (shiftreg2[fractbits+4]) && (!(&shiftreg2[fractbits+3:fractbits+2])); // shiftreg2 < -4
	ol1n = (shiftreg2[fractbits+4]) && (!(&shiftreg2[fractbits+3:fractbits])); // shiftreg2 < -1
	olquartern = (shiftreg2[fractbits+4]) && (!(&shiftreg2[fractbits+3:fractbits-2])); // shiftreg2 < -1/4

// Juiste grenswaarden selecteren afhankelijk van het volgnummer van de integrator
// Selecting the proper limiting thresholds depending on the number of the integrator
	case (shiftcount)
		3'b010: // feitelijk data van de eerste integrator in shiftreg2, actually data from the first integrator in shiftreg2
		begin
			olp = olquarterp;
			ollp = {6'sb000000, {(fractbits-2) {1'b1}}}; // bijna 1/4 met 4 bits voor en fractbits (29) achter de komma, almost 1/4 with 4 bits before and fractbits (29) after the binary point
			oln = olquartern;
			olln = {6'sb111111, {(fractbits-2) {1'b0}}}; // -1/4 met 4 bits voor en 29 achter de komma, -1/4 with 4 bits before and 29 after the binary point
		end
		3'b011: // feitelijk data van de tweede integrator in shiftreg2, actually data from the second integrator in shiftreg2
		begin
			olp = ol1p;
			ollp = {4'sb0000, {(fractbits) {1'b1}}}; // bijna 1 met 4 bits voor en 29 achter de komma, almost 1 with 4 bits before and 29 after the binary point
			oln = ol1n;
			olln = {4'sb1111, {(fractbits) {1'b0}}}; // -1 met 4 bits voor en 29 achter de komma, -1 with 4 bits before and 29 after the binary point
		end
		3'b100: // feitelijk data van de derde integrator in shiftreg2, actually data from the third integrator in shiftreg2
		begin
			olp = ol4p;
			ollp = {4'sb0011, {(fractbits) {1'b1}}}; // bijna 4 met 4 bits voor en 29 achter de komma, almost 4 with 4 bits before and 29 after the binary point
			oln = ol4n;
			olln = {4'sb1100, {(fractbits) {1'b0}}}; // -4 met 4 bits voor en 29 achter de komma, -4 with 4 bits before and 29 after the binary point
		end
		default: // vierde of vijfde integrator, fourth or fifth integrator
		begin
			olp = ol8p;
			ollp = {4'sb0111, {(fractbits) {1'b1}}}; // bijna 8 met 4 bits voor en 29 achter de komma, almost 8 with 4 bits before and 29 after the binary point
			oln = ol8n;
			olln = {4'sb1000, {(fractbits) {1'b0}}}; // -8 met 4 bits voor en 29 achter de komma, -8 with 4 bits before and 29 after the binary point
		end
	endcase
	case ({olp, oln})
		2'b10: shiftreg2lim = ollp;
		2'b01: shiftreg2lim = olln;
		default: shiftreg2lim = shiftreg2[fractbits+3:0];
	endcase
	iclip = olp || oln; // pulsje van 1 sdclock5-cyclus bij integratoroversturing, pulse of 1 sdclock5 cycle when an integrator clips
end
pwm pwm1(q, rotifen, sdclock, y); // Het feitelijke herbemonsteren gebeurt hier. The actual resampling happens here.
endmodule
`default_nettype wire

